#!/usr/bin/python2
# -*- coding: utf-8 -*-
# nop
from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import tempfile
import traceback
import subprocess
import array
import cairo
import ftplib
import random
import string
import datetime
from math import pi
from os import remove
from os.path import dirname
from os.path import isfile
from os.path import join
from os.path import expanduser
from threading import Thread
from yaml import dump
from yaml import load
# from wand.image import Image
from synapseqrcodes import synapse_qrcode
from .svgparse import SvgParser
from .references import MultipleFiles
from .references import MissingFile
from .imgconverter import convert_png_to_bmp
try:
    from gi import require_version
except ImportError:
    print("ERROR: pygobject is not installed.")
    raise
try:
    require_version("Gtk", "3.0")
except ValueError:
    print("ERROR: gtk3 development files are not installed.")
    raise
except AttributeError:
    print("pygobject version too old.")
    raise
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from logging import getLogger
from logging import basicConfig
from logging import INFO
# from six import iteritems
from six import StringIO
# from .labelprintingop import PrintingApp
from synapseqrcodes import verify_parameter
from .jscript_utils import shift_and_format_items
from .pw_utils import PasswordGenerator
from .utils import get_path
from .references import ReferenceConfig
from .references import reference_config_file
from .labelconfig import label_config_file
from .utils import svg_dir
try:
    from synapselabels import __version__
    assert __version__
except ImportError:
    __version__ = "UNKNOWN"

LABELS_GENERATED_FILE = join(expanduser('~'), 'LABELS_GENERATED')
LABELS_PRINTED_FILE = join(expanduser('~'), 'LABELS_PRINTED')
log = getLogger('synapselabels')
dnd_targets = [('text/plain', Gtk.TargetFlags.OTHER_WIDGET, 0), ]
default_unit = 'm'
DEMO_WITH_HEADER = (
    "date code\tethernet mac address\tethernet mac address last 6\t"
    "snap mac address\twifi mac address\t"
    "snap node address\t"
    "default password for user access\tdefault password for wifi access\t"
    "sequential code\tbatch size\tproduction order number\t"
    "modem meid\tmodem imei\tsim iccid\tHW version\n"

    "1744\t001C2CAAAAAA\tAAAAAA\t"
    "001C2C0000DDDDDD\t001C2CAAAAA0\t"
    "DDDDDD\t"
    "AawawawAawawawawa(79\tDawawawDawawawawa(79\t"
    "1371\t2000\t1234567890\t"
    "A1000042F1866B\t356278079904497\t89148000001470824524\tRevA\n"

    "1745\t001C2CBBBBBB\tBBBBBB\t"
    "001C2C0000EEEEEE\t001C2CBBBBB0\t"
    "AAAAAA\t"
    "BawawawBawawawawa(79\tEawawawEawawawawa(79\t"
    "1372\t2000\t1234567890\t"
    "A1000042F1866C\t356278079904498\t89148000001470824525\tRevB\n"

    "1746\t001C2CCCCCCC\tCCCCCCC\t"
    "001C2C0000FFFFFF\t001C2CCCCCC0\t"
    "FFFFFF\t"
    "CawawawCawawawawa(79\tFawawawFawawawawa(79\t"
    "1273\t2000\t1234567890\t"
    "A1000042F1866D\t356278079904499\t89148000001470824526\tRevC\n"
    )

fonts_dir = join(dirname(__file__), 'fonts')

"""
def get_font_downloads():
    font_downloads = []
    for font_name, font_file_name in (
        ('ARIAL', 'Arial.ttf'),
        # ('ARIALB', 'Arial_Bold.ttf'),
        # ('TERBU', 'Trebuchet_MS.ttf'),
        # ('TERBUB', 'Trebuchet_MS_Bold.ttf'),
            ):
        font_file_path = join(
            fonts_dir,
            font_file_name
            )
        if isfile(font_file_path):
            with open(font_file_path, 'rb') as font_stream:
                font_binary = font_stream.read().replace(
                    '\x1b', '\x1b\x1b'
                    )
            font_download = (
                'd ' + 'TTF' + ';' + font_name + '\r' +
                '\x1b.' + font_binary + '\x1b.'
                )
            font_downloads.append(font_download)
    return font_downloads
"""


def save_label(label):
    if 'created' not in label:
        label['created'] = timestamp()
    with open(LABELS_GENERATED_FILE, 'a') as yaml_file:
        yaml_file.write('- ' + dump(label, width=10000))


class ImageFileMissing(Exception):
    pass


def list_of_dicts_to_table(data_log):
    params = set()
    for record in data_log:
        for key, value in record.items():
            params.add(key)
    params = sorted(list(params))
    print('params:', params)
    table = '\t'.join(params)
    for record in data_log:
        line_items = []
        for param in params:
            line_items.append(record[param] if param in record else '')
        table += '\n' + '\t'.join(line_items)
    return table


def timestamp():
    return datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')


def hexadecimal_increment(hex_string):
    if len(hex_string) == 0:
        return '1'
    last = 1 + int(hex_string[-1:], 16)
    if last <= 15:
        return hex_string[:-1].upper() + ('%X' % last)
    else:
        return hexadecimal_increment(hex_string[:-1]) + '0'


def decimal_increment(dec_string):
    if len(dec_string) == 0:
        return '1'
    last = 1 + int(dec_string[-1:])
    if last <= 9:
        return dec_string[:-1] + str(last)
    else:
        return decimal_increment(dec_string[:-1]) + '0'


def start_file(filepath):
    if sys.platform.startswith('darwin'):
        subprocess.call(('open', filepath))
    elif os.name == 'nt':
        os.startfile(filepath)
    elif os.name == 'posix':
        subprocess.call(('xdg-open', filepath))


glade_file = get_path('labelsgui.glade')
src1 = get_path('names.txt')
src2 = get_path('places.txt')


def gtk_input(
    parent,
    parameter,
    record,
    records,
    default='',
    autoincrement_option=False,
    fixed_option=False,
    autogenerate_option=False,
        ):
    """
    Display a dialog with a text entry.
    Returns the text, or None if canceled.
    """
    message = (
        "Please Enter or Scan\n<b><big>%s</big></b>\n(%d of %d)" %
        (parameter, record + 1, records)
        )
    d = Gtk.MessageDialog(
        parent,
        Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
        Gtk.MessageType.QUESTION,
        Gtk.ButtonsType.OK_CANCEL,
        # message
        )
    vbox = d.get_message_area()
    d.set_markup(message)
    entry = Gtk.Entry()
    entry.set_text(default)
    entry.show()
    if autoincrement_option:
        auto_inc = Gtk.CheckButton(label='Autoincrement')
        auto_inc.show()
        vbox.pack_end(auto_inc, True, True, 0)
    if autogenerate_option:
        auto_gen = Gtk.CheckButton(label='Autogenerate')
        auto_gen.show()
        vbox.pack_end(auto_gen, True, True, 0)
    if fixed_option:
        fixed = Gtk.CheckButton(label='Fixed')
        fixed.show()
        vbox.pack_end(fixed, True, True, 0)
    vbox.pack_end(entry, True, True, 0)
    entry.connect('activate', lambda _: d.response(Gtk.ResponseType.OK))
    d.set_default_response(Gtk.ResponseType.OK)
    r = d.run()
    text = entry.get_text().decode('utf8')
    d.destroy()
    if r == Gtk.ResponseType.OK:
        return (
            text,
            auto_gen.get_active() if autogenerate_option else
            auto_inc.get_active() if autoincrement_option else
            fixed.get_active() if fixed_option else
            False
            )


def dots(size, unit, dpi):
    if size.startswith('pt'):
        size = float(size[2:],) / 72.0
    else:
        size = float(size)
        if unit == 'm':
            size = size / 25.4  # convert mm to inch
        elif unit != 'i':
            raise ValueError(
                'invalid unit: %s' % repr(unit)
                )
    return dpi * size


def pil2cairo(im):
    """Transform a PIL Image into a Cairo ImageSurface."""

    assert sys.byteorder == 'little', 'We don\'t support big endian'
    if im.mode != 'RGBA':
        im = im.convert('RGBA')

    s = im.tobytes('raw', 'BGRA')
    a = array.array('B', s)
    dest = cairo.ImageSurface(cairo.FORMAT_ARGB32, im.size[0], im.size[1])
    ctx = cairo.Context(dest)
    non_premult_src_wo_alpha = cairo.ImageSurface.create_for_data(
        a, cairo.FORMAT_RGB24, im.size[0], im.size[1])
    non_premult_src_alpha = cairo.ImageSurface.create_for_data(
        a, cairo.FORMAT_ARGB32, im.size[0], im.size[1])
    ctx.set_source_surface(non_premult_src_wo_alpha)
    ctx.mask_surface(non_premult_src_alpha)
    return dest


def get_selected_data(treeview):
        # print("print_selected_label")
        (model, pathlist) = treeview.get_selection().get_selected_rows()
        for path in pathlist:
            tree_iter = model.get_iter(path)
            for index, column in enumerate(
                    treeview.get_columns()):
                return model.get_value(tree_iter, index)


def on_drag_data_get(treeview, context, selection, info, timestamp):
    """ Extract data from the source of the DnD operation.

    Serialize iterators of selected tasks in format
    <iter>,<iter>,...,<iter> and set it as parameter of DND """

    selected_data = get_selected_data(treeview)
    selection.set(selection.get_target(), 0, selected_data)


def set_treeview_columns(treeview, columns):
    for i in range(len(columns)):
        # cellrenderer to render the text
        cell = Gtk.CellRendererText(editable=False)
        # cell.set_property("editable", True)
        # the text in the first column should be in boldface
        # if i == 0:
        #    cell.props.weight_set = True
        #    cell.props.weight = Pango.Weight.BOLD
        # the column is created
        col = Gtk.TreeViewColumn(
            columns[i], cell, text=i)
        # and it is appended to the treeview
        treeview.append_column(col)


def set_treeview_from_clipboard(treeview, with_header=True):
    set_treeview_from_text_data(
        treeview,
        Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).wait_for_text(),
        with_header
        )


def set_treeview_from_text_data(treeview, text_data, with_header=True):
    listmodel = None
    if text_data is not None:
        # print("clipboard text:", repr(text_data))
        for line in text_data.splitlines():
            # print("line:", repr(line))
            texts = line.split('\t')
            if listmodel is None:
                for column in treeview.get_columns():
                    treeview.remove_column(column)
                args = [GObject.TYPE_STRING] * len(texts)
                # print(args)
                listmodel = Gtk.ListStore(*args)
                # listmodel = Gtk.ListStore(str, str, str)
                if with_header:
                    columns = texts
                else:
                    columns = [
                        "Column{}".format(index) for
                        index in
                        range(0, len(texts))
                        ]
                    # print("appending:", texts)
                    listmodel.append(texts)
            else:
                # print("appending:", texts)
                listmodel.append(texts)
        if listmodel is not None:
            treeview.set_model(listmodel)
            set_treeview_columns(treeview, columns)
    else:
        print("No text on the clipboard.")


def set_treeview_from_texts(treeview, texts):
    # treeview.set_headers_visible(False)
    # print('set treeview from: ', repr(texts))
    listmodel = Gtk.ListStore(GObject.TYPE_STRING)
    if texts is not None:
        for line in texts:
            listmodel.append([line])
    treeview.set_model(listmodel)


def fill_combo(combo, data):
    store = Gtk.ListStore(GObject.TYPE_STRING)
    for line in data:
        store.append([line])
    combo.set_model(store)
    combo.set_active(0)
    # And here's the new stuff:
    # cell = Gtk.CellRendererText()
    # combo.pack_start(cell, True)
    # combo.add_attribute(cell, "text", 0)


def get_treeview_column(treeview, x_pos):
    for column in treeview.get_columns():
        if column.get_x_offset() <= x_pos:
            if x_pos <= (column.get_x_offset() + column.get_width()):
                return column


def drag_data_received(
        widget, context, x_pos, y_pos, selection, info, timestamp):
    data = selection.get_data()
    print('drag_data_received data:', data)
    if data:
        column = get_treeview_column(widget, x_pos)
        if column:
            print(
                "old column title:", repr(column.get_title()),
                "new column title:", repr(data)
                )
            column.set_title(data)


class LabelsGui(SvgParser, ReferenceConfig):
    """
    This is the E20 GUI
    """
    HEADER_HEIGHT = 10 * 72 / 25.4
    HEADER_GAP = 3 * 72 / 25.4

    def __init__(self):
        """
        Build GUI
        """
        SvgParser.__init__(self, label_config_file)
        ReferenceConfig.__init__(self, reference_config_file)
        self.template = None
        self.autoincremented_params = {}
        self.fixed_params = {}
        self.autogenerated_params = []
        self.scanned_qr_code = {}
        self.sales_level_part_number = None
        self.dpi = 300
        # Build GUI from Glade file
        self.builder = Gtk.Builder()
        self.builder.add_from_file(glade_file)
        # Get objects
        go = self.builder.get_object
        self.button_view_document = go('buttonViewDocument')
        self.button_view_document.connect(
            "clicked",
            self.view_document,
            )
        self.window = go('applicationwindow1')
        self.window.set_title(
            'Synapse Labels ' + __version__
            )
        self.textview = go('textview')
        self.drawingarea = go('drawingarea')
        self.drawingarea.connect("draw", self.on_draw)
        self.treeview_data_table = go('treeviewDataTable')
        self.treeview_data_table.get_selection().connect(
            "changed",
            self.data_tree_selection_changed,
            "dummy"
            )
        self.treeview_data_table.enable_model_drag_dest(
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
            )
        self.treeview_data_table.connect(
            'drag-data-received',
            drag_data_received,
            )
        # self.combobox = go('comboboxtext1')
        # self.mac_num = go('labelMacNum')
        self.entry_date_code = go('entryDateCode')
        self.entry_records = go('entryRecords')
        self.entry_records.connect(
            "changed",
            self.entry_records_changed
            )
        self.button_today = go('buttonToday')
        self.entry_copies = go('entryCopies')
        self.button_today.connect(
            "clicked",
            self.button_today_clicked,
            "today button data"
            )
        self.button_generate = go('buttonGenerate')
        self.button_generate.connect(
            "clicked",
            self.button_generate_clicked,
            )
        self.button_upgrade = go('buttonUpgrade')
        self.button_upgrade.connect(
            "clicked",
            self.button_upgrade_clicked,
            )
        self.button1 = go('button1')
        self.button1.connect(
            "clicked",
            self.button_1_clicked,
            )
        self.button2 = go('button2')
        self.button2.connect(
            "clicked",
            self.button_2_clicked,
            )
        self.button3 = go('button3')
        self.button3.connect(
            "clicked",
            self.button_3_clicked,
            )
        self.button4 = go('button4')
        self.button4.connect(
            "clicked",
            self.button_4_clicked,
            )
        self.label_number_of_records = go('labelNumberOfRecords')
        self.button_demo = go('buttonDemo')
        self.button_demo.connect(
            "clicked",
            self.button_demo_clicked,
            )
        self.button_generated = go('buttonGenerated')
        self.button_generated.connect(
            "clicked",
            self.button_generated_clicked,
            )
        self.button_printed = go('buttonPrinted')
        self.button_printed.connect(
            "clicked",
            self.button_printed_clicked,
            )
        self.button_paste_with_header = go('buttonPasteWithHeader')
        self.button_paste_with_header.connect(
            "clicked",
            self.button_paste_with_header_clicked,
            "cool button 2"
            )
        self.button_print_selected_labels = go('buttonPrintSelectedLabels')
        self.button_print_selected_labels.connect(
            "clicked",
            self.print_selected_labels,
            )
        self.button_print_every_label = go('buttonPrintEveryLabel')
        self.button_print_every_label.connect(
            "clicked",
            self.print_every_label,
            )

        self.button_sheet_selected_labels = go('buttonSheetSelectedLabels')
        self.button_sheet_selected_labels.connect(
            "clicked",
            self.sheet_selected_labels,
            )
        self.button_sheet_every_label = go('buttonSheetEveryLabel')
        self.button_sheet_every_label.connect(
            "clicked",
            self.sheet_every_label,
            )

        self.cb_printers = go('comboboxPrinters')
        self.cb_printers.connect(
            "changed",
            self.cb_printers_changed,
            "dummy"
            )
        self.cb_part_numbers = go('comboboxPartNumbers')
        self.treeview_params = go('treeviewParams')
        set_treeview_columns(self.treeview_params, ['Params'])
        self.treeview_params.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
            )
        self.treeview_params.connect('drag_data_get', on_drag_data_get)
        self.button_view_qr_code_standard = go('buttonViewQrCodeStandard')
        self.button_view_qr_code_standard.connect(
            "clicked",
            self.view_qr_code_standard,
            )
        self.button_view_spd = go('buttonViewSpd')
        self.button_view_spd.connect(
            "clicked",
            self.view_spd,
            )
        self.entry_spd = go('entrySpd')
        self.entry_document = go('entryDocument')
        self.entry_label_template = go('entryLabelTemplate')
        self.entry_label_type = go('entryLabelType')
        self.entry_qr_code_standard = go('entryQrCodeStandard')
        self.box_qr_code_standard = go('boxQrCodeStandard')
        self.box_preview = go('boxPreview')
        self.box_upgrade = go('boxUpgrade')
        fill_combo(self.cb_printers, sorted(self.printers.keys()))
        self.cb_printers.set_active(-1)
        fill_combo(self.cb_part_numbers, sorted(self.part_numbers.keys()))
        self.cb_part_numbers.set_active(-1)
        self.cb_documents = go('comboboxDocuments')
        self.cb_documents.connect(
            "changed",
            self.cb_documents_selection_changed,
            "dummy"
            )
        self.cb_part_numbers.connect(
            "changed",
            self.cb_part_numbers_selection_changed,
            "dummy"
            )
        self.treeview_params.get_selection().connect(
            "changed",
            self.params_tree_selection_changed,
            "dummy"
            )
        self.set_date_code_to_today()
        # Connect signals
        self.builder.connect_signals(self)
        self.cb_part_numbers_selection_changed(self.cb_part_numbers)
        # Configure interface
        self.window.connect('delete-event', lambda x, y: Gtk.main_quit())
        # GObject.timeout_add(1000, self.update_mac_pool)
        # Everything is ready
        self.window.show()
        self.password_generator = PasswordGenerator(src1, src2)
        thread = Thread(target=self.check_update)
        thread.daemon = True
        thread.start()

    def button_upgrade_clicked(self, widget, data=None):
        thread = Thread(target=self.upgrade)
        thread.daemon = True
        thread.start()
        self.box_upgrade.set_visible(False)

    def upgrade(self):
        args = [
            'pip2',
            'install',
            '-U',
            '--disable-pip-version-check',
            '--extra-index-url',
            'https://update.synapse-wireless.com/pypi',
            'synapselabels',
            ]
        try:
            print(subprocess.check_output(args))
            print('upgrade completed successfully, please restart')
        except subprocess.CalledProcessError:
            # sudo is not working if operator is logged in!
            # if os.name == 'posix':
            #     print('upgrade failed, retrying upgrade with gksudo')
            #     args = ['gksudo', ' '.join(args)]
            #     try:
            #         print(subprocess.check_output(args))
            #         print('upgrade completed successfully, please restart')
            #         return
            #     except subprocess.CalledProcessError:
            #         print('upgrade failed, please upgrade manually')
            #         self.box_upgrade.set_visible(True)
            print('upgrade failed, please upgrade manually')
            self.box_upgrade.set_visible(True)

    def check_update(self):
        self.box_upgrade.set_visible(False)
        self.window.set_title(
            'Synapse Labels ' + __version__ + ' (version check in progress)'
            )
        args = [
            'pip2',
            'list',
            '--disable-pip-version-check',
            '--extra-index-url',
            'https://update.synapse-wireless.com/pypi',
            '-o',
            '--format=legacy'
            ]
        try:
            updatables = subprocess.check_output(args)
            for updatable in updatables.splitlines():
                if updatable.startswith('synapselabels '):
                    try:
                        latest_index = updatable.index('Latest: ')
                        latest = updatable[latest_index + 8:]
                        latest = latest.split(' ', 1)
                        latest = latest[0]
                        print('latest:', latest)
                        self.box_upgrade.set_visible(True)
                        self.window.set_title(
                            'Synapse Labels ' +
                            __version__ +
                            ' (latest available is %s)' % latest
                            )
                    except Exception:
                        print('cannot parse:', updatable)
                    return
            self.window.set_title(
                'Synapse Labels ' + __version__ + ' (this is latest available)'
                )
        except subprocess.CalledProcessError:
            self.window.set_title(
                'Synapse Labels ' + __version__ + ' (version check failed)'
                )

    def get_copies(self):
        try:
            return int(self.entry_copies.get_text())
        except Exception:
            return 1

    def print_selected_labels(self, widget, data=None):
        self.print_labels(
            self.get_selected_labels(),
            self.entry_label_type.get_text(),
            self.get_copies(),
            )

    def sheet_selected_labels(self, widget, data=None):
        self.sheet_labels(
            self.get_selected_labels(),
            self.entry_label_type.get_text(),
            self.get_copies(),
            )

    def get_selected_labels(self):
        # print("print_selected_label")
        selection = self.treeview_data_table.get_selection()
        # print('selection:', selection)
        (model, pathlist) = selection.get_selected_rows()
        labels = []
        for path in pathlist:
            # print('path:', path)
            tree_iter = model.get_iter(path)
            # print('tree_iter:', tree_iter)
            label_data = {}
            for param in self.get_parameters():
                value = None
                for index, column in enumerate(
                        self.treeview_data_table.get_columns()):
                    if column.get_title() == param:
                        value = model.get_value(
                            tree_iter,
                            index,
                            )
                        break
                # print('value:', value)
                if value is None:
                    if param == 'date code':
                        entry_date_code_text = self.entry_date_code.get_text()
                        if len(entry_date_code_text) == 4:
                            value = entry_date_code_text
                if value:
                    label_data[param] = value
                else:
                    print(repr(param), 'parameter is missing')
                    return
            labels.append(label_data)
        return labels

    def get_label_from_row(self, row):
        label_data = {}
        for param in self.get_parameters():
            value = None
            for index, column in enumerate(
                    self.treeview_data_table.get_columns()):
                if column.get_title() == param:
                    value = row[index]
                    break
            # print('value:', value)
            if value:
                label_data[param] = value
            else:
                print(repr(param), 'parameter is missing')
                return
        return label_data

    def get_all_labels(self):
        labels = []
        for row in self.treeview_data_table.get_model():
            label_data = self.get_label_from_row(row)
            if label_data:
                labels.append(label_data)
        return labels

    def print_every_label(self, widget, data=None):
        self.print_labels(
            self.get_all_labels(),
            self.entry_label_type.get_text(),
            self.get_copies(),
            )

    def sheet_every_label(self, widget, data=None):
        self.sheet_labels(
            self.get_all_labels(),
            self.entry_label_type.get_text(),
            self.get_copies()
            )

    def button_1_clicked(self, widget, data=None):
        self.entry_copies.set_text('1')

    def button_2_clicked(self, widget, data=None):
        self.entry_copies.set_text('2')

    def button_3_clicked(self, widget, data=None):
        self.entry_copies.set_text('3')

    def button_4_clicked(self, widget, data=None):
        self.entry_copies.set_text('4')

    def get_number_of_records(self):
        model = self.treeview_data_table.get_model()
        return len(model) if model else 0

    def adjust_number_records(self):
        number_of_records = self.get_number_of_records()
        self.label_number_of_records.set_text(str(number_of_records))
        self.button_print_every_label.set_sensitive(
            number_of_records > 0 and
            self.get_selected_printer_config() is not None
            )
        self.button_sheet_every_label.set_sensitive(number_of_records > 0)

    def button_paste_clicked(self, widget, data=None):
        # print("button_paste_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=False)
        self.adjust_number_records()

    def button_paste_with_header_clicked(self, widget, data=None):
        # print("button_paste_with_header_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=True)
        self.adjust_number_records()

    def button_demo_clicked(self, widget, data=None):
        Gtk.Clipboard.get(
            Gdk.SELECTION_CLIPBOARD
            ).set_text(
                DEMO_WITH_HEADER,
                -1
                )
        self.button_paste_with_header.emit("clicked")

    def button_generated_clicked(self, widget, data=None):
        try:
            with open(LABELS_GENERATED_FILE, 'r') as yaml_file:
                data_log = load(yaml_file.read())
            records = self.get_records()
            if records is not None:
                data_log = data_log[-records:]
            # print('generated labels:\n', data_log)
            self.set_records(data_log)
        except IOError:
            pass  # silently pass, because log file does not exist yet

    def button_printed_clicked(self, widget, data=None):
        try:
            with open(LABELS_PRINTED_FILE, 'r') as yaml_file:
                data_log = load(yaml_file.read())
            records = self.get_records()
            if records is not None:
                data_log = data_log[-records:]
            # print('printed labels:\n', data_log)
            self.set_records(data_log)
        except IOError:
            pass  # silently pass, because log file does not exist yet

    def set_records(self, data_log):
        set_treeview_from_text_data(
            self.treeview_data_table,
            list_of_dicts_to_table(data_log),
            with_header=True,
            )
        self.adjust_number_records()

    def data_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        csr = widget.count_selected_rows()
        # print("count_selected_rows:", csr)
        self.button_print_selected_labels.set_sensitive(
            csr is not None and
            (csr > 0) and
            self.get_selected_printer_config() is not None
            )
        self.button_sheet_selected_labels.set_sensitive(csr > 0)
        self.update_label_preview()

    def cb_printers_changed(self, widget, data=None):
        number_of_records = self.get_number_of_records()
        self.button_print_every_label.set_sensitive(
            number_of_records > 0 and self.get_selected_printer_config()
            )
        csr = self.treeview_data_table.get_selection().count_selected_rows()
        self.button_print_selected_labels.set_sensitive(
            csr is not None and
            (csr > 0) and
            self.get_selected_printer_config() is not None
            )

    def update_label_preview(self):
        if self.template in self.templates:
            self.box_preview.set_visible(True)
        else:
            self.box_preview.set_visible(False)
            return
        selected_labels = self.get_selected_labels()
        if selected_labels is not None and len(selected_labels) == 1:
            size_x, size_y = self.get_label_size()

            self.selected_label = selected_labels[0]
            # print('preview_label:', self.selected_label)
            self.selected_label_file = self.get_label_file(self.selected_label)
            # Gnm.textview_set_text(
            #     self.textview.set_text,
            #     self.selected_label_file
            #     )
            self.textview.get_buffer().set_text(self.selected_label_file)
            # self.preview_label_file(self.get_label_file(label))
            self.drawingarea.set_size_request(
                self.dpi * size_x / 25.4,
                self.dpi * size_y / 25.4
                )
            self.drawingarea.queue_draw()
            # self.window.reshow_with_initial_size()
        else:
            self.textview.get_buffer().set_text('')
            self.drawingarea.set_size_request(-1, -1)
            self.selected_label = None
            self.label_file = None
            # self.window.reshow_with_initial_size()

    # def preview_label_file(self, label_file):
    #     print('preview_label_file', label_file)
    #     for line in label_file.splitlines():
    #         if line.startswith('T'):
    #             print ("preview", line)
    #             control, text = line.split(';', 1)
    #             control = control[2:]
    #             pos_x, pos_y, rotate, font, size = control.split(',')
    #             print ("text", pos_x, pos_y, rotate, font, size, text)

    def on_draw(self, wid, cr):
        if self.template not in self.templates:
            return
        # frame it!
        size_x, size_y = self.get_label_size()
        cr.set_line_width(1)
        cr.rectangle(
            0,
            0,
            self.dpi * size_x / 25.4,
            self.dpi * size_y / 25.4,
            )
        cr.set_source_rgb(255, 255, 255)
        cr.fill()
        # cr.set_source_rgb(0, 0, 0)
        # cr.stroke()
        self.draw_label_file(
            (
                self.selected_label_file
                if hasattr(self, 'selected_label_file') else ''
                ),
            cr,
            self.dpi
            )

    def draw_label_file(self, label_file, cr, dpi):
        # cr.set_font_size(dpi * 6.0 / 72)
        unit = default_unit
        for line in label_file.splitlines():
            if line.startswith('m'):
                unit = line[2:]  # m or i
                if unit == 'i':
                    # print('unit is inch')
                    pass
                elif unit == 'm':
                    # print('unit is mm')
                    pass
                else:
                    print('unsupported line:', line)
            elif line.startswith('J'):
                if len(line) != 1:
                    print('unsupported line:', line)
            elif line.startswith('S'):
                # size TBD
                pass
            elif line.startswith('O'):
                # TBD
                pass
            elif line.startswith('A'):
                # TBD
                pass
            elif line.startswith('T'):
                # print("preview", line)
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, font, size = control.split(',')
                # print("text", pos_x, pos_y, rotate, font, size, text)

                size = dots(size, unit, dpi)
                cr.set_font_size(size)
                if font == '3':
                    cr.select_font_face(
                        "Swiss 721",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_NORMAL,
                        )
                elif font == '596':
                    cr.select_font_face(
                        "Monospace 821",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_NORMAL,
                        )
                elif font == '5':
                    cr.select_font_face(
                        "Swiss 721",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_BOLD,
                        )
                # elif font == '10':
                #     cr.select_font_face(
                #         "Arial",
                #         cairo.FONT_SLANT_NORMAL,
                #         cairo.FONT_WEIGHT_NORMAL,
                #         )
                # elif font == '11':
                #     cr.select_font_face(
                #         "Arial",
                #         cairo.FONT_SLANT_NORMAL,
                #         cairo.FONT_WEIGHT_BOLD,
                #         )
                # elif font == '12':
                #     cr.select_font_face(
                #         "Trebuchet MS",
                #         cairo.FONT_SLANT_NORMAL,
                #         cairo.FONT_WEIGHT_NORMAL,
                #         )
                # elif font == '13':
                #     cr.select_font_face(
                #         "Trebuchet MS",
                #         cairo.FONT_SLANT_NORMAL,
                #         cairo.FONT_WEIGHT_BOLD,
                #         )
                else:
                    print('unsupported font number:', font)
                cr.set_source_rgb(0, 0, 0)
                if rotate == '0':
                    # textents = cr.text_extents(text)
                    # print('text:', text, 'textents:', textents)
                    # print(
                    #     'pos_x_dots:', dots(pos_x, unit, dpi),
                    #     'pos_y_dots:', dots(pos_y, unit, dpi)
                    #     )
                    cr.move_to(
                        dots(pos_x, unit, dpi),  # - textents[0],
                        dots(pos_y, unit, dpi),
                        )
                    cr.show_text(text)
                elif rotate == '90':
                    # textents = cr.text_extents(text)
                    cr.move_to(
                        dots(pos_x, unit, dpi),
                        dots(pos_y, unit, dpi),  # + textents[0]
                        )
                    cr.save()
                    cr.rotate(1.5 * pi)
                    cr.show_text(text)
                    cr.restore()
                elif rotate == '270' or rotate == '-90':
                    # textents = cr.text_extents(text)
                    cr.move_to(
                        dots(pos_x, unit, dpi),
                        dots(pos_y, unit, dpi),  # - textents[0]
                        )
                    cr.save()
                    cr.rotate(0.5 * pi)
                    cr.show_text(text)
                    cr.restore()
                elif rotate == '180':
                    # textents = cr.text_extents(text)
                    cr.move_to(
                        dots(pos_x, unit, dpi),  # + textents[0],
                        dots(pos_y, unit, dpi),
                        )
                    cr.save()
                    cr.rotate(pi)
                    cr.show_text(text)
                    cr.restore()
                else:
                    print('unsupported rotate:', rotate)
            elif line.startswith('G'):
                # print("preview", line)
                # G 0.05,0.64,0;R:1.9,0.33,0.0066666
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate = control.split(',')
                function, function_params = text.split(':', 1)
                if function == 'R':
                    width, height, thickness = function_params.split(',', 2)
                    cr.set_line_width(dots(thickness, unit, dpi))
                    cr.rectangle(
                        dots(pos_x, unit, dpi),
                        dots(pos_y, unit, dpi),
                        dots(width, unit, dpi),
                        dots(height, unit, dpi),
                        )
                    cr.set_source_rgb(0, 0, 0)
                    cr.stroke()
                else:
                    print('unsupported line:', line)
            elif line.startswith('B'):
                # print("preview", line)
                # B 1.39,0.08,0,QRCODE+EL{error level}+MODEL2,.0133333;{qrcode}
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, barcode, size = control.split(',', 4)
                barcode_opts = barcode.split('+')
                if barcode_opts[0] == 'QRCODE':
                    qr_model = 1
                    error_correction = 'L'
                    for opt in barcode_opts:
                        if opt == 'QRCODE':
                            pass
                        elif opt == 'MODEL1':
                            qr_model = 1
                        elif opt == 'MODEL2':
                            qr_model = 2
                        elif opt == 'ELL':
                            error_correction = 'L'
                        elif opt == 'ELM':
                            error_correction = 'M'
                        elif opt == 'ELQ':
                            error_correction = 'Q'
                        elif opt == 'ELH':
                            error_correction = 'H'
                        else:
                            print('unsupported qrcode option:', opt)
                    box_size = int(0.5 + dots(size, unit, dpi))
                    if qr_model != 2:
                        print('unsupported QR Code model:', qr_model)
                    try:
                        with tempfile.NamedTemporaryFile(
                            suffix='.png',
                            prefix='qr',
                            delete=False
                                ) as qr_file:
                            qr_file_name = qr_file.name
                        args = [
                            'qrencode',
                            '-o', qr_file_name,
                            '-l', error_correction,
                            '-m', '0',
                            '-s', str(box_size),
                            text
                            ]
                        # print("args:", args)
                        subprocess.check_call(args)
                        surface = cairo.ImageSurface.create_from_png(
                            qr_file_name
                            )
                    except Exception:
                        traceback.print_exc(file=sys.stdout)
                    finally:
                        try:
                            if isfile(qr_file_name):
                                remove(qr_file_name)
                                # print('removed temporary file:', qr_file_name)
                        except Exception:
                            print('cannot remove temporary file:', qr_file_name)
                            traceback.print_exc(file=sys.stdout)
                    # cr.set_antialias(cairo.ANTIALIAS_NONE)
                    cr.set_source_surface(
                        surface,
                        dots(pos_x, unit, dpi),
                        dots(pos_y, unit, dpi)
                        )
                    cr.paint()
                    # print('cr', dir(cr))
                else:
                    print('unsupported line:', line)
            elif line.startswith('I'):
                # I [:name;]x,y,r[,mx,my,GOODBADn][,a];name
                # implement:
                # I x,y,r;name
                control, image_name = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate = control.split(',', 2)
                # print('image_name:', repr(image_name))
                image_file_name = self.get_image_file(image_name)
                # print('image_file_name:', repr(image_file_name))
                surface = cairo.ImageSurface.create_from_png(
                    get_path(image_file_name)
                    )
                cr.set_source_surface(
                    surface,
                    dots(pos_x, unit, dpi),
                    dots(pos_y, unit, dpi)
                    )
                cr.paint()
            else:
                print('unsupported line:', line)

    def get_image_file(self, name):
        for file_name, short_name in self.image_names.items():
            if name == short_name:
                local_file = join(svg_dir, file_name)
                if not isfile(local_file):
                    raise ImageFileMissing(local_file)
                return local_file
        if name in self.tempfiles:
            return self.tempfiles[name]
        raise ValueError('Unrecognized image name: %s' % name)

    def params_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        # csr = widget.count_selected_rows()
        # print("params_tree_selection_changed count_selected_rows:", csr)
        pass

    def get_parameters(self):
        try:
            return self.templates[self.template]['parameters']
        except Exception:
            # silent fail, because it is possible to have a label without param
            return []

    def get_label_type(self):
        return self.label_types[
            self.templates[self.template]['label type']
            ]

    def get_label_size(self):
        label_type = self.get_label_type()
        return label_type['width'], label_type['height']

    def cb_part_numbers_selection_changed(self, widget, data=None):
        # print("cb_part_numbers_selection_changed - %s" % data)
        # reset a few entries
        self.document_name = None
        self.entry_document.set_text('')
        self.template = None
        self.entry_label_template.set_text('')
        self.entry_label_type.set_text('')
        self.qr_code_standard = None
        self.entry_qr_code_standard.set_text('')
        self.box_qr_code_standard.set_visible(False)
        set_treeview_from_texts(
            self.treeview_params,
            [],
            )
        # find out the selected part number
        tree_iter = widget.get_active_iter()
        if tree_iter is not None:
            sales_level_part_number = widget.get_model()[tree_iter][0]
            # TBD
            if sales_level_part_number in self.part_numbers:
                self.sales_level_part_number = sales_level_part_number
                part_number_data = self.part_numbers[
                    self.sales_level_part_number
                    ]
                self.cb_part_numbers.set_tooltip_text(
                    part_number_data['Description']
                    if 'Description' in part_number_data else ''
                    )
                if 'spd' in part_number_data:
                    spd = part_number_data['spd']
                    # print('spd:', spd)
                    if spd in self.spds:
                        self.spd = spd
                        spd_data = self.spds[spd]
                        self.items = (
                            spd_data['items'] if 'items' in spd_data else {}
                            )
                        spd_entry = spd
                        if 'Revision' in spd_data:
                            spd_entry += '_' + spd_data['Revision']
                        self.entry_spd.set_text(spd_entry)
                        self.entry_spd.set_tooltip_text(
                            spd_data['Description']
                            if 'Description' in spd_data else ''
                            )
                    else:
                        self.items = {}
                else:
                    self.spd = None
                    self.entry_spd.set_text('')
                    self.items = (
                        part_number_data['items']
                        if 'items' in part_number_data
                        else {}
                        )
                fill_combo(
                    self.cb_documents,
                    sorted(self.items.keys())
                    )
        # self.update_label_preview()

    def cb_documents_selection_changed(self, widget, data=None):
        tree_iter = widget.get_active_iter()
        if tree_iter is not None:
            self.document_title = widget.get_model()[tree_iter][0]
            item = self.items[self.document_title]
            self.document_name = (
                item['part number'] if 'part number' in item else ''
                )
            self.entry_document.set_text(
                self.document_name + '_' + item['Revision']
                if 'Revision' in item else self.document_name
                )
            self.entry_document.set_tooltip_text(
                item['Description'] if 'Description' in item else ''
                )
            if 'qr code standard' in item:
                self.qr_code_standard = item['qr code standard']
                self.entry_qr_code_standard.set_text(self.qr_code_standard)
                self.box_qr_code_standard.set_visible(True)
            else:
                self.qr_code_standard = None
                self.entry_qr_code_standard.set_text('')
                self.box_qr_code_standard.set_visible(False)
            self.template = (
                item['template'] if 'template' in item else self.document_name
                )
            try:
                document_file = self.get_file(self.document_name)
                self.button_view_document.set_sensitive(True)
                self.button_view_document.set_tooltip_text(document_file)
            except (MultipleFiles, MissingFile):
                self.button_view_document.set_sensitive(False)
                self.button_view_document.set_tooltip_text('')
            self.entry_label_type.set_text(
                self.templates[self.template]['label type']
                if self.template in self.templates else ''
                )
            self.entry_label_type.set_tooltip_text(
                dump(self.templates[self.template]['label type'], width=50)
                if self.template in self.templates else ''
                )
            self.entry_label_template.set_text(
                self.template if self.template
                else ''
                )
            set_treeview_from_texts(
                self.treeview_params,
                self.get_parameters()
                )
            self.update_label_preview()

    def button_today_clicked(self, widget, data=None):
        # print("button_today_clicked - %s" % data)
        self.set_date_code_to_today()

    def set_date_code_to_today(self):
        today = datetime.date.today()
        todays_week = today.strftime("%W")
        todays_year = today.strftime("%Y")
        todays_date_code = str(todays_year[-2:]) + str(todays_week)
        self.entry_date_code.set_text(todays_date_code)

    def get_label_file(self, label_params, x_shift=None, y_shift=None):
        for param in self.get_parameters():
            if param not in label_params:
                raise ValueError('label parameter %s is missing' % repr(param))
        label_format_params = {
            'sales level part number': self.sales_level_part_number,
            }
        label_format_params.update(label_params)
        if 'date code' in label_params:
            label_format_params['YY'] = label_params['date code'][:2]
            label_format_params['WW'] = label_params['date code'][2:]
        if 'snap mac address' in label_params:
            label_format_params[
                'snap node address'
                ] = label_params['snap mac address'][-6:]
        if 'ethernet mac address' in label_params:
            label_format_params[
                'ethernet mac address last 6'
                ] = label_params['ethernet mac address'][-6:]
        # print('label_format_params:', label_format_params)
        if self.template not in self.templates:
            print('not implemented label type: %s' % self.template)
            return None
        svg_config_data = self.templates[self.template]
        if 'method' in svg_config_data:
            method = svg_config_data['method']
            if method == 'bitmap':
                return shift_and_format_items(
                    self.get_label_bitmap(
                        self.template,
                        label_format_params
                        )['items'],
                    x_shift=x_shift,
                    y_shift=y_shift,
                    label_format_params=label_format_params
                    )
        return shift_and_format_items(
            self.get_label(self.template)['items'],
            x_shift=x_shift,
            y_shift=y_shift,
            label_format_params=label_format_params
            )

    def get_image_downloads(self, label_file):
        image_downloads = []
        for line in label_file.splitlines():
            if line.startswith('I'):
                print('image line:', line)
                control, image_name = line.split(';')
                image_file = self.get_image_file(image_name)
                bmp_file = image_file[:-3] + 'bmp'
                if not isfile(bmp_file):
                    with tempfile.NamedTemporaryFile(
                        prefix='',
                        suffix='.bmp',
                        delete=False
                            ) as temp_file:
                        bmp_file = temp_file.name
                    convert_png_to_bmp(
                        image_file,
                        bmp_file
                        )
                    # start_file(bmp_file)
                    print('bmp_file:', bmp_file)
                    with open(bmp_file, 'rb') as bmp_stream:
                        bmp_binary = bmp_stream.read().replace(
                            '\x1b', '\x1b\x1b'
                            )
                    remove(bmp_file)
                else:
                    with open(bmp_file, 'rb') as bmp_stream:
                        bmp_binary = bmp_stream.read().replace(
                            '\x1b', '\x1b\x1b'
                            )
                image_download = (
                    'd ' + 'BMP' + ';' + image_name + ' \r' +
                    '\x1b.' + bmp_binary + '\x1b.'
                    )
                print(
                    'image_name:', image_name,
                    'image_file_path:', image_file,
                    'length:', len(image_download),
                    )
                image_downloads.append(image_download)
        return image_downloads

    def print_label_file_to(self, label_file, printer_config):
        # create random file name
        filename = ''.join(random.choice(string.digits) for _ in range(9))
        # print("random file name:", filename)
        if 'ftp' in printer_config:
            ftp_config = printer_config['ftp']
            ftp_connection = None
            try:
                ftp_connection = ftplib.FTP(
                    ftp_config['ip address']
                    )
                ftp_connection.login(
                    ftp_config['user name'],
                    ftp_config['password']
                    )
                ftp_connection.cwd("execute")
                ftp_connection.storbinary(
                    'STOR ' + filename,
                    StringIO(label_file)
                    )
            except Exception:
                print("FTP Error")
                traceback.print_exc(file=sys.stdout)
            finally:
                if ftp_connection is not None:
                    ftp_connection.close()

    def get_selected_printer_name(self):
        tree_iter = self.cb_printers.get_active_iter()
        if tree_iter is not None:
            return self.cb_printers.get_model()[tree_iter][0]

    def get_selected_printer_config(self):
        selected_printer_name = self.get_selected_printer_name()
        if selected_printer_name is not None:
            return self.printers[selected_printer_name]

    def print_label_file(self, label_file, label_records):
        # print('print_label_file:\n"""\n%s"""' % label_file)
        # font_downloads = get_font_downloads()
        image_downloads = self.get_image_downloads(label_file)
        # label_file = '\n'.join(font_downloads) + label_file
        if len(image_downloads) > 0:
            label_file = '\r\n'.join(image_downloads) + '\r\n' + label_file
        selected_printer_config = self.get_selected_printer_config()
        if selected_printer_config is not None:
            try:
                self.print_label_file_to(label_file, selected_printer_config)
                self.save_print_records(label_records)
            except Exception:
                print("Could not print labels")
        else:
            print("No printer is selected")

    def save_print_records(self, labels):
        # labels = self.get_all_labels()
        time_before = None
        for label in labels:
            label['sales level part number'] = self.sales_level_part_number
            while True:
                time_now = timestamp()
                if time_now != time_before:
                    time_before = time_now
                    break
            label['created'] = time_now
            selected_printer_name = self.get_selected_printer_name()
            if selected_printer_name is not None:
                label['printer'] = selected_printer_name
            if self.template is not None:
                label['template'] = self.template
            with open(LABELS_PRINTED_FILE, 'a') as yaml_file:
                yaml_file.write('- ' + dump(label, width=10000))

    def print_labels(self, labels, label_type_name, copies=1):
        print('print labels:', labels)
        label_type = self.label_types[label_type_name]
        x_shift = 0
        lpr = label_type['lpr']
        total_width = label_type['width']
        if lpr > 1:
            total_width += (lpr - 1) * label_type['hrepeat']
        # print('lpr:', lpr)
        label_file = None
        label_records = []
        if labels is not None:
            for label in labels:
                for copy in range(copies):
                    label_records.append(label)
                    current_label_file = self.get_label_file(
                        label,
                        (x_shift * label_type['hrepeat'])
                        if x_shift else None
                        )
                    if label_file is None:
                        # label_file = 'M l fnt;arial\n'
                        label_file = 'm m\r\n'
                        label_file += 'J\r\n'
                        # label_file = 'F 10;Arial\n'
                        # label_file = 'F 11;ARIALB\n'
                        # label_file = 'F 12;TREBU\n'
                        # label_file = 'F 13;TREBUB\n'
                        label_file += (
                            'S l1;0,0,' +
                            str(label_type['height']) + ',' +
                            str(label_type['vrepeat']) + ',' +
                            str(total_width) +
                            '\r\n'
                            )
                        label_file += 'O R\r\n'
                    label_file += current_label_file
                    x_shift += 1
                    if x_shift >= lpr:
                        label_file += 'A 1\r\n'
                        self.print_label_file(label_file, label_records)
                        label_file = None
                        label_records = []
                        x_shift = 0
        if label_file is not None:
            label_file += 'A 1\r\n'
            self.print_label_file(label_file, label_records)
            label_file = None
            label_records = []
            x_shift = 0

    def button_generate_clicked(self, widget, data=None):
        # print("button_generate_clicked - %s" % data)
        set_treeview_from_text_data(
            self.treeview_data_table,
            self.generate_records(),
            with_header=True,
            )
        self.adjust_number_records()
        self.save_records()

    def save_records(self):
        labels = self.get_all_labels()
        time_before = None
        for label in labels:
            label['sales level part number'] = self.sales_level_part_number
            while True:
                time_now = timestamp()
                if time_now != time_before:
                    time_before = time_now
                    break
            label['created'] = time_now
            save_label(label)

    def get_records(self):
        try:
            return int(self.entry_records.get_text())
        except Exception:
            return None

    def entry_records_changed(self, widget, data=None):
        records = self.get_records()
        self.button_generate.set_sensitive(records is not None and records > 0)

    def generate_records(self):
        try:
            records = int(self.entry_records.get_text())
        except ValueError:
            records = 0
        # print('records', records)
        parameters = self.get_parameters()
        retval = '\t'.join(parameters) + '\n'
        # reset the generators
        self.autoincremented_params = {}
        self.fixed_params = {}
        self.autogenerated_params = []
        self.scanned_qr_code = {}
        for record in range(records):
            more = record < (records - 1)
            retval += '\t'.join(
                [
                    self.generate_parameter(
                        parameter,
                        more,
                        record,
                        records,
                        ) for parameter in parameters
                    ]
                ) + '\n'
        return retval

    def generate_parameter(self, parameter, more, record, records):
        if parameter in [
            'date code',
                ]:
            # prompt operator for optionally fixed parameter
            if parameter in self.fixed_params:
                new_param = self.fixed_params[parameter]
            else:
                while True:
                    new_param, fixed = gtk_input(
                        self.window,
                        parameter,
                        record,
                        records,
                        default=self.entry_date_code.get_text(),
                        fixed_option=True,
                        )
                    if new_param is None:
                        return None
                    new_param = new_param.strip()
                    try:
                        self.scanned_qr_code = synapse_qrcode(new_param)
                        new_param = self.scanned_qr_code[parameter]
                    except Exception:
                        # silently pass,
                        # if it was not a Synapse QR code it is not a problem
                        pass
                    try:
                        verify_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
                if fixed:
                    self.fixed_params[parameter] = new_param
            return new_param
        elif parameter in [
            'HW version',
                ]:
            # prompt operator for optionally fixed parameter
            if parameter in self.fixed_params:
                new_param = self.fixed_params[parameter]
            else:
                while True:
                    new_param, fixed = gtk_input(
                        self.window,
                        parameter,
                        record,
                        records,
                        fixed_option=True,
                        )
                    if new_param is None:
                        return None
                    new_param = new_param.strip()
                    try:
                        self.scanned_qr_code = synapse_qrcode(new_param)
                        new_param = self.scanned_qr_code[parameter]
                    except Exception:
                        # silently pass,
                        # if it was not a Synapse QR code it is not a problem
                        pass
                    try:
                        verify_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
                if fixed:
                    self.fixed_params[parameter] = new_param
            return new_param
        elif parameter in [
            'production order number',
            'batch size',
                ]:
            # prompt operator for forcefully fixed parameter
            if parameter in self.fixed_params:
                new_param = self.fixed_params[parameter]
            else:
                while True:
                    new_param, fixed = gtk_input(
                        self.window,
                        parameter,
                        record,
                        records,
                        default=(
                            self.scanned_qr_code[parameter]
                            if parameter in self.scanned_qr_code else ''
                            )
                        )
                    if new_param is None:
                        return None
                    new_param = new_param.strip()
                    try:
                        self.scanned_qr_code = synapse_qrcode(new_param)
                        new_param = self.scanned_qr_code[parameter]
                    except Exception:
                        # silently pass,
                        # if it was not a Synapse QR code it is not a problem
                        pass
                    new_param = new_param.upper()
                    try:
                        verify_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
                self.fixed_params[parameter] = new_param
            return new_param
        elif parameter in [
            'default password for user access',
            'default password for wifi access'
                ]:
            # automatically generate parameter, no prompt
            if parameter in self.autogenerated_params:
                return self.password_generator.generate_password()
            else:
                while True:
                    new_param, auto_gen = gtk_input(
                        self.window,
                        parameter,
                        record,
                        records,
                        autogenerate_option=True,
                        default=(
                            self.scanned_qr_code[parameter]
                            if parameter in self.scanned_qr_code else ''
                            )
                        )
                    if new_param is None:
                        return None
                    if auto_gen:
                        self.autogenerated_params.append(parameter)
                        return self.password_generator.generate_password()
                    new_param = new_param.strip()
                    try:
                        self.scanned_qr_code = synapse_qrcode(new_param)
                        new_param = self.scanned_qr_code[parameter]
                    except Exception:
                        # silently pass,
                        # if it was not a Synapse QR code it is not a problem
                        pass
                    try:
                        verify_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
            return new_param
        elif parameter in [
            'ethernet mac address',
            'snap mac address',
            'wifi mac address',
            'snap node address',
            'modem meid',
                ]:
            # prompt operator for optionally autoincremented
            # hexadecimal parameter
            if parameter in self.autoincremented_params:
                new_param = hexadecimal_increment(
                    self.autoincremented_params[parameter]
                    )
                self.autoincremented_params[parameter] = new_param
            else:
                while True:
                    new_param, auto_inc = gtk_input(
                        self.window,
                        parameter,
                        record,
                        records,
                        autoincrement_option=more,
                        default=(
                            self.scanned_qr_code[parameter]
                            if parameter in self.scanned_qr_code else ''
                            )
                        )
                    if new_param is None:
                        return None
                    new_param = new_param.strip()
                    print('new_param:', new_param)
                    try:
                        self.scanned_qr_code = synapse_qrcode(new_param)
                        print('self.scanned_qr_code:', self.scanned_qr_code)
                        new_param = self.scanned_qr_code[parameter]
                        print(parameter, new_param)
                    except Exception:
                        # silently pass,
                        # if it was not a Synapse QR code it is not a problem
                        traceback.print_exc(file=sys.stdout)
                        pass
                    new_param = new_param.upper()
                    try:
                        verify_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
                if auto_inc:
                    self.autoincremented_params[parameter] = new_param
            return new_param
        elif parameter in [
            'sim iccid',
            'sequential code',
                ]:
            # prompt operator for optionally autoincremented
            # decimal parameter
            if parameter in self.autoincremented_params:
                new_param = decimal_increment(
                    self.autoincremented_params[parameter]
                    )
                self.autoincremented_params[parameter] = new_param
            else:
                while True:
                    new_param, auto_inc = gtk_input(
                        self.window,
                        parameter,
                        record,
                        records,
                        autoincrement_option=more,
                        default=(
                            self.scanned_qr_code[parameter]
                            if parameter in self.scanned_qr_code else ''
                            )
                        )
                    if new_param is None:
                        return None
                    new_param = new_param.strip()
                    try:
                        self.scanned_qr_code = synapse_qrcode(new_param)
                        new_param = self.scanned_qr_code[parameter]
                    except Exception:
                        # silently pass,
                        # if it was not a Synapse QR code it is not a problem
                        pass
                    try:
                        verify_parameter(parameter, new_param)
                        break
                    except ValueError as error:
                        print(error)
                if auto_inc:
                    self.autoincremented_params[parameter] = new_param
            return new_param
        else:
            # TBD pop-up info dialog instead
            raise ValueError(
                "cannot generate parameter: %s" %
                repr(parameter)
                )

    def sheet_labels(self, labels, label_type_name, copies=1):
        # typical paper and margin sizes
        # letter size: 8.5 x 11 in.
        # printable area on letter: 8.2 x 10.6 in.
        # top margin .22 in.
        # bottom margin: .18 in
        # sides margin .15 in
        label_type = self.label_types[label_type_name]
        label_x_mm = label_type['width']
        label_y_mm = label_type['height']
        paper_x_mm = 8.5 * 25.4
        paper_y_mm = 11 * 25.4
        side_margin_mm = 0.15 * 25.4
        top_margin_mm = 0.22 * 25.4
        bottom_margin_mm = 0.15 * 25.4
        printable_x_mm = paper_x_mm - side_margin_mm - side_margin_mm
        printable_y_mm = paper_y_mm - top_margin_mm - bottom_margin_mm
        labels_x = int(printable_x_mm / label_x_mm)
        labels_y = int(printable_y_mm / label_y_mm)
        # print(
        #     'labels_x:', labels_x, 'labels_y:', labels_y,
        #     'len:', len(labels)
        #     )
        # labels_page = labels_x * labels_y
        doc_dir = GLib.get_user_special_dir(
            GLib.UserDirectory.DIRECTORY_DOCUMENTS
            )
        if doc_dir is None:
            doc_dir = GLib.get_home_dir()
        pdf_file = join(doc_dir, 'gtk-demo.pdf')
        # print('pdf_file:', pdf_file)
        pdfdpi = 300.0
        ps = cairo.PDFSurface(
            pdf_file,
            pdfdpi * paper_x_mm / 25.4,
            pdfdpi * paper_y_mm / 25.4
            )
        cr = cairo.Context(ps)
        x_pos = 0
        y_pos = 0
        for label in labels:
            for copy in range(copies):
                # print('x_pos:', x_pos, 'y_pos:', y_pos)
                x_shift_mm = side_margin_mm + x_pos * label_x_mm
                y_shift_mm = top_margin_mm + y_pos * label_y_mm
                label_file = self.get_label_file(
                    label,
                    x_shift=x_shift_mm,
                    y_shift=y_shift_mm
                    )
                # frame it!
                cr.set_line_width(1)
                cr.rectangle(
                    pdfdpi * x_shift_mm / 25.4,
                    pdfdpi * y_shift_mm / 25.4,
                    pdfdpi * label_x_mm / 25.4,
                    pdfdpi * label_y_mm / 25.4,
                    )
                # cr.set_source_rgb(255, 255, 255)
                # cr.fill()
                cr.set_source_rgb(0, 0, 0)
                cr.stroke()
                try:
                    self.draw_label_file(label_file, cr, pdfdpi)
                except Exception:
                    traceback.print_exc(file=sys.stdout)
                x_pos += 1
                if x_pos >= labels_x:
                    x_pos = 0
                    y_pos += 1
                    if y_pos >= labels_y:
                        y_pos = 0
                        cr.show_page()
        if x_pos > 0 or y_pos > 0:
            cr.show_page()
        ps.finish()
        if isfile(pdf_file):
            start_file(pdf_file)

    def view_spd(self, widget, data=None):
        if self.spd is not None:
            start_file(self.get_file(self.spd))

    def view_document(self, widget, data=None):
        if self.document_name is not None:
            start_file(self.get_file(self.document_name))

    def view_qr_code_standard(self, widget, data=None):
        if self.qr_code_standard is not None:
            start_file(self.get_file(self.qr_code_standard))


def synapse_labels():
    basicConfig(level=INFO)
    gui = LabelsGui()
    assert gui
    Gtk.main()


if __name__ == "__main__":
    synapse_labels()
